#include <z3++.h>
#include <cstdio>
#include <map>
#include <vector>

constexpr uint32_t SIZE = 5;

using state_t = std::map<std::string, z3::expr>;

z3::expr make_var(const std::string& name, z3::context& ctx, int idx, uint32_t size=SIZE)
{
    return ctx.bv_const((name + "_" + std::to_string(idx)).c_str(), size);
}

state_t create_state(z3::context& ctx, int idx)
{
    state_t the_state;
    the_state.emplace("cnt", make_var("cnt", ctx, idx));
    the_state.emplace("inc", make_var("inc", ctx, idx));
    the_state.emplace("rst", make_var("rst", ctx, idx, 1));
    
    the_state.emplace("add", the_state.at("cnt") + the_state.at("inc"));
    the_state.emplace("lt",  z3::ule(the_state.at("inc"), 5));
    the_state.emplace("eq",  the_state.at("add") == 15);
    the_state.emplace("mux", z3::ite(the_state.at("rst") == 1,
                                     ctx.bv_val(0, SIZE), the_state.at("add")));
    return the_state;
}

void environment_constraints(const state_t& the_state, z3::solver& solver)
{
    solver.add(the_state.at("lt"));
}

void transition_constraints(const state_t& curr_state, const state_t& prev_state, z3::solver& solver)
{
    solver.add(curr_state.at("cnt") == prev_state.at("mux"));
}

void initial_constraints(const state_t& the_state, z3::solver& solver)
{
    solver.add(the_state.at("cnt") == 0);
}

void bad_state_properties(const state_t& the_state, z3::solver& solver)
{
    z3::expr_vector bad_conds(solver.ctx());
    bad_conds.push_back(the_state.at("eq"));
    solver.add(z3::mk_or(bad_conds));
}


int main()
{
    z3::context ctx;
    z3::solver solver(ctx);
    std::vector<state_t> states;

    for (int i = 0; i < 3; i++)
    {
        state_t curr_state = create_state(ctx, i);
        states.push_back(curr_state);
        
        if (i == 0)
        {
            initial_constraints(curr_state, solver);
        }
        else 
        {
            transition_constraints(curr_state, states[i-1], solver);
        }
        environment_constraints(curr_state, solver);
        
        solver.push();
        bad_state_properties(curr_state, solver);

        auto res = solver.check();
        if (res == z3::sat)
        {
            z3::model model = solver.get_model();
            for (int i = 0; i < states.size(); i += 1)
            {
                printf("state_%d:\n", i);
                for (auto& it: states[i])
                {
                    printf("%s = %ld\n", it.first.c_str(), model.eval(it.second).as_uint64());
                }
            }
        }
        solver.pop();
        
    }

    

    return 0;
}
